#include <gtk/gtk.h>
#include <string.h>
#include "../Misc/my_dbug.h"
#include "../Misc/mem_management.h"
#include "../Misc/garbage_collection.h"
#include "../Syntaxtree/syntaxtree.h"
#include "eval_global.h"

static GList* var_list;
static GList* replace_list;


/* @returns TRUE if the pattern matches
 *          FALSE if not 
 */
static gboolean  
checkSingleClause(Expr_node_ptr params, Expr_node_ptr args)
{
  gboolean res = TRUE;

  DBUG_ENTER("checkSingleClause");
  DBUG_ASSERT(params != NULL);
  DBUG_ASSERT(args   != NULL);

  //printf_expr_node(params);
  switch (params->nodetype) {
  case e_wildcard:
    DBUG_PRINT("pm", ("e_wildcard"));
    res = (args != NULL);
    break;

  case e_constant:
    //case e_nil_const:
    DBUG_PRINT("pm", ("e_constant"));
    res = areMLValuesEqual(args, params);
    break;

    /*
      case e_nil_const:
      DBUG_PRINT("pm", (""));
      res = (args->nodetype == params->nodetype);
      break;
    */
  case e_listkeeper:
    DBUG_PRINT("pm", ("e_listkeeper"));
    if (args->nodetype == params->nodetype) {
      if ( (LIST_EXPR_NODE(params) != NULL) && 
	   (LIST_EXPR_NODE(args)   != NULL) ) {
	res = checkSingleClause(LIST_EXPR_NODE(params), 
				LIST_EXPR_NODE(args));
      } else {
	res = ( (LIST_EXPR_NODE(params) == NULL) && 
		(LIST_EXPR_NODE(args)   == NULL) );
      } //if
    } else {
      res = FALSE;
    } //if same nodetype
    break;

  case e_list:
    DBUG_PRINT("pm", ("e_list"));
    res = ( (args->nodetype == params->nodetype)  &&
	    checkSingleClause(LISTITEM_EXPR_NODE(params), 
			      LISTITEM_EXPR_NODE(args)) );
    if (res && (LISTNEXT_EXPR_NODE(params) != NULL) 
	    && (LISTNEXT_EXPR_NODE(args)   != NULL) ) {
      res = checkSingleClause(LISTNEXT_EXPR_NODE(params), 
			      LISTNEXT_EXPR_NODE(args));
    } else {
      res = res && ( (LISTNEXT_EXPR_NODE(params) == NULL) && 
		     (LISTNEXT_EXPR_NODE(args)   == NULL) );
    } //if-else
    break;

  case e_param_list:
    DBUG_PRINT("pm", ("e_param_list"));
    res = ( (args->nodetype == e_arg_list)  &&
	    checkSingleClause(PARAM_EXPR_NODE(params), 
			      ARG_EXPR_NODE(args)) );

    if (res && (NEXT_EXPR_NODE(params) != NULL )
	&& (NEXT_EXPR_NODE(args) != NULL)) {
      res = checkSingleClause(NEXT_EXPR_NODE(params), 
			      NEXT_EXPR_NODE(args));
    }
    break;

  case e_variable:
    DBUG_PRINT("pm", ("e_variable"));
    if (args == NULL) {
      res = FALSE;
    } else {
      DBUG_ASSERT(args->nodetype != e_list);
      var_list     = g_list_prepend(var_list,     params->strrep);
      replace_list = g_list_prepend(replace_list, args);
      res = TRUE;
    }
    break;

  case e_constructor:
    DBUG_PRINT("pm", ("e_constructor"));
    if ( (args->nodetype != params->nodetype) 
	 || !args->strrep || !params->strrep
	 || (0 != strcmp(args->strrep, params->strrep)) ) {
      res = FALSE;
    } else {
      if (PARAMLIST_EXPR_NODE(params) != NULL) {
	res = checkSingleClause(PARAMLIST_EXPR_NODE(params), 
				ARGLIST_EXPR_NODE(args));
      } else {
	res = TRUE;
      }
    }
    break;

  case e_prepend:
    DBUG_PRINT("pm", ("e_prepend"));
    if ((args->nodetype == e_listkeeper) && (LIST_EXPR_NODE(args) != NULL)) {
      struct _Mlval_struct mlval;
      Expr_node_ptr tmp_listkeeper;

      mlval.val.mllist = NULL;
      mlval.lineno = -1;
      mlval.colno  = -1;
      tmp_listkeeper = 
	create_expr_listkeeper( mlval, 
			        LISTNEXT_EXPR_NODE(LIST_EXPR_NODE(args)) );
      gcAddItem(tmp_listkeeper);
      res = checkSingleClause( LEFT_OPERAND_EXPR_NODE(params), 
			       LISTITEM_EXPR_NODE(LIST_EXPR_NODE(args)) )  &&
            checkSingleClause( RIGHT_OPERAND_EXPR_NODE(params), 
			       tmp_listkeeper);

    } else if (args->nodetype == e_prepend) {
      res = checkSingleClause( LEFT_OPERAND_EXPR_NODE(params), 
			       LEFT_OPERAND_EXPR_NODE(args)    )  &&
            checkSingleClause( RIGHT_OPERAND_EXPR_NODE(params), 
			       RIGHT_OPERAND_EXPR_NODE(args)   );

    } else {
      res = FALSE;
    } //if
    break;

  default:
    DBUG_PRINT("pm", ("default"));
    printf_expr_node(params);
    res = FALSE; //only to avoid a compiler-warning
    DBUG_ASSERT(0);
  } //switch

  DBUG_PRINT("eval", ("res == %d", res));
  DBUG_RETURN(res);
} //checkSingleClause


/* @returns TRUE if a matching was found
 *          FALSE if no pattern matches (runtimeerror) */
gboolean  
patternMatching(Fun_node_ptr pFun, Expr_node_ptr args, 
		Expr_node_ptr* refBody_clause, 
		GList** refVariable_list, GList** refReplacement_list)
{
  gboolean success, res;
  GList* params_list, *body_list;

  DBUG_ENTER("patternMatching");

  var_list     = NULL;
  replace_list = NULL;
  success      = FALSE;
  params_list     = pFun->paramslist;
  body_list       = pFun->bodylist;

  while (!success && (params_list != NULL) && (body_list != NULL)) {
    my_g_list_free_except_items(var_list);
    my_g_list_free_except_items(replace_list);
    var_list     = NULL;
    replace_list = NULL;
    if (params_list->data == NULL) {
      DBUG_ASSERT(args == NULL);
      success = TRUE;
    } else {
      success = checkSingleClause(params_list->data, args);
    }
    params_list  = g_list_next(params_list);
    if (!success) {
      body_list = g_list_next(body_list);
    }
  } //while

  if (success) {
    *refVariable_list    = var_list;
    *refReplacement_list = replace_list;
    *refBody_clause      = body_list->data;
    res = TRUE;
  } else {
    gchar* tmpStr;
    my_g_list_free_except_items(var_list);
    my_g_list_free_except_items(replace_list);
    var_list     = NULL;
    replace_list = NULL;
    tmpStr = mtstrconcat(2, "no matching pattern found for function ", 
			    pFun->fun_id);
    runtime_error(tmpStr);
    mtfree(tmpStr);
    res = FALSE;
  }

  DBUG_PRINT("eval", ("res == %d", res));
  DBUG_RETURN(res);
} //patternMatching
