/** This module implements a single garbage collection.
  *
  * This garbage collection can keep an not fixed number of pointers to
  * dynamical allocated (using 'mtalloc') memory, which have to be freed
  * at the same time (using 'mtfree'). ('mtalloc' and 'mtfree' are functions
  * implemented in "mem_management.c")  
  *
  * Befor the first use it has to be initialized by 'gcInitialize' or
  * 'gcInitializeBySize' and after last use it has to be finalized by
  * 'gcFinalize'.
  * Other functions are 'gcAddItem' and 'gcFreeItems'.
  */
#include <gtk/gtk.h>
#include "my_dbug.h"
#include "mem_management.h"
#include "garbage_collection.h"


typedef struct {
  void** items;
  gint    nr_items;
  gint    max_nr_items;
} Garbage_collection_struct, *Garbage_collection_ptr;

static Garbage_collection_struct gc;


/** Initailizes the garbage collection for beeing able to keep at least a 
  * given number of items.
  * If the number items grows its size will be increased.
  */
void gcInitializeBySize(gint initial_size)
{
  DBUG_ENTER("gcInitializeBySize");

  gc.items        = g_malloc(initial_size * sizeof(void*));
  gc.nr_items     = 0;
  gc.max_nr_items = initial_size;

  DBUG_VOID_RETURN;
} //gcInitializeBySize


/** Initailizes the garbage collection for beeing able to keep at least a 
  * suggested number of items.
  * If the number items grows its size will be increased.
  */
void gcInitialize(void)
{
  const gint gc_initial_size = 128;

  DBUG_ENTER("gcInitialize");

  gcInitializeBySize(gc_initial_size);
  
  DBUG_VOID_RETURN;
} //gcInitialize


/** Frees all items and free all recourses used by the garbage collection.
  */
void gcFinalize(void)
{
  DBUG_ENTER("gcFinalize");

  gcFreeItems();
  
  g_free(gc.items);
  gc.items        = NULL;
  gc.nr_items     = 0;
  gc.max_nr_items = 0;

  DBUG_VOID_RETURN;
} //gcFinalize


/** Frees all stored items by using 'mtfree'.
  */
void gcFreeItems(void)
{
  gint i;

  DBUG_ENTER("gcFreeItems");

  for (i=0; i < gc.nr_items; i++) {
    mtfree(gc.items[i]);
  } //for i
  gc.nr_items     = 0;

  DBUG_VOID_RETURN;
} //gcFreeItems


/** Adds an item and increases the number of storeable items, if necessary.
  */
void gcAddItem(void* ptr)
{
  gint new_max;

  DBUG_ENTER("gcAddItem");
  DBUG_ASSERT(gc.items != NULL);

  if (ptr != NULL) {
    if (gc.nr_items == gc.max_nr_items) {
      //increase size of garbage collection
      new_max  = gc.max_nr_items + (gc.max_nr_items / 3);
      gc.items = g_realloc(gc.items, new_max * sizeof(void*));
      gc.max_nr_items = new_max;
    }
    gc.items[gc.nr_items] = ptr;
    gc.nr_items++;
  } else {
    DBUG_PRINT("", ("you tried to add NULL to the garbage collection"));
  } //if ptr != NULL

  DBUG_VOID_RETURN;
} //gcAddItem
