/** This module covers the standard memory allocating functions to do 
  * additional profiling for debugging purpose and implements additional
  * usefull functions.
  *
  * The module can keep a fixed number of (pointers to) memory segments with
  * the following additional information: size in bytes, filename and line
  * in sourcefile, where it was allocated.
  * Additionally the total memory in use (in bytes) and the total number
  * of segments is provided.
  */
#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include "my_dbug.h"
#include "utilities.h"
#include "mem_management.h"
#include "../GTK/drawable_expr.h" //debug only

#define MAX_FILENAMESIZE  ((size_t)32)
#define MAX_SLOTS     100000

static struct {
  gpointer address;  //NULL means this slot in unused
  gsize    size;
  gchar    file[MAX_FILENAMESIZE]; //file where this slot was occupied
  gint     line;     //line within the file above.
} mtmem[MAX_SLOTS];
static gint totalMemInUse = 0;
static gint totalSegmentsInUse = 0;


/** Initializes the mem_management, but a finalization is not needed!
  */
void 
mtmemory_init(void)
{
  gint i;

  DBUG_ENTER("mtmemory_init");

  for(i=0; i < MAX_SLOTS; i++) {
    mtmem[i].address = NULL;
    mtmem[i].size = 0;
    mtmem[i].file[0] = '\0';
    mtmem[i].line = 0;
  }
  totalMemInUse      = 0;
  totalSegmentsInUse = 0;

  DBUG_VOID_RETURN;
} //mtmemory_init


/** Allocates memory of the given size and does some memory profiling.
  * @return a pointer to fresh allocated memory.
  */
gpointer
mtmalloc_dbug_on(gsize size, gchar* file, gint line)
{
  gint i;
  gpointer res;

  DBUG_ENTER("mtmalloc_dbug_on");

  i = 0;
  while((mtmem[i].address != NULL) && (i < MAX_SLOTS)) {
    i++;
  }
  if (i == MAX_SLOTS) {
    puts("mem_management-error: to many mtmallocs");
    gtk_exit(1);
  }

  res = g_malloc0(size);
  totalMemInUse += size;
  totalSegmentsInUse++;

  mtmem[i].address = res;
  mtmem[i].size    = size;
  strncpy(mtmem[i].file, file, MAX_FILENAMESIZE-1);
  mtmem[i].file[MAX_FILENAMESIZE-1] = '\0'; //guatantee terminated string
  mtmem[i].line    = line;

  DBUG_RETURN(res);
} //mtmalloc_dbug_on


/** Rellocates memory of the given size and does some memory profiling.
  * @return a pointer to memory.
  */
gpointer
mtrealloc_dbug_on(gpointer old_ptr, gsize size, gchar* file, gint line)
{
  gint i;
  gpointer res;

  DBUG_ENTER("mtrealloc_dbug_on");

  i = 0;
  while((mtmem[i].address != old_ptr) && (i < MAX_SLOTS)) {
    i++;
  }
  if (i == MAX_SLOTS) {
    res =mtmalloc_dbug_on(size, file, line);
  } else {
    res = g_realloc(old_ptr, size);
    totalMemInUse += size - mtmem[i].size;
    //totalSegmentsInUse is still the same

    mtmem[i].address = res;
    mtmem[i].size    = size;
    strncpy(mtmem[i].file, file, MAX_FILENAMESIZE-1);
    mtmem[i].file[MAX_FILENAMESIZE-1] = '\0'; //guatantee terminated string
    mtmem[i].line    = line;
  }

  DBUG_RETURN(res);
} //mtrealloc_dbug_on


/** Frees memory which was allocated by mtalloc.
  */
void 
mtfree_dbug_on(gpointer ptr, gchar* file, gint line)
{
  gint i;

  DBUG_ENTER("mtfree_dbug_on");

  if (ptr == NULL) {
    DBUG_PRINT("free", ("pointer is already NULL"));
    //printf("pointer is already NULL\n");
  } else {
    i = 0;
    while(((mtmem[i].address == NULL) || (mtmem[i].address != ptr))
          && (i < MAX_SLOTS)) i++;
    if (i == MAX_SLOTS) {
      printf("mem_management-error: pointer was not allocated by "
	     "mtmalloc (%p)\n"
	     "- point of call: file '%s', line %d\n"
	     "- pointer as int %d;  pointer as double %f\n"
	     "- pointer as string '%s'\n", 
	     ptr, file, line, *(gint*)ptr, *(gdouble*)ptr, (gchar*)ptr);
      gtk_exit(1);
    }

    totalMemInUse -= mtmem[i].size;
    totalSegmentsInUse--;

    mtmem[i].address = NULL;
    mtmem[i].size = 0;
    mtmem[i].file[0] = '\0';
    mtmem[i].line = 0;
    G_FREENULL(ptr);
  } //if ptr != NULL

  DBUG_VOID_RETURN;
} //mtfree_dbug_on


void 
mtmultifree(gint nrArgs, ...)
{
  gint i;
  gpointer tmp;
  va_list ap; /* ap points to each argument */

  DBUG_ENTER("mtmultifree");

  va_start(ap, nrArgs); /* initialization for va_arg */
  i = 1;
  while (i <= nrArgs) {
    tmp = va_arg(ap, gpointer); /* get actual argument */
    //printf("mtmultifree: arg %i is '%s' (%p)\n", i, tmp, tmp);
    MTFREENULL(tmp);
    i++;
  }
  va_end(ap); /* finalization for va_arg */

  DBUG_VOID_RETURN;
} //mtmultifree


/** Shows the total memory in use and total segments in use.
  */
void 
mtshowtotals_dbug_on(gchar* tag)
{
  DBUG_ENTER("mtshowtotals_dbug_on");

  printf("total mem in use: %6i bytes;  %5i slots;  (%s)\n", 
         totalMemInUse, totalSegmentsInUse, tag);

  DBUG_VOID_RETURN;
} //mtshowtotals_dbug_on


/** Prints all used slots and its profiled information.
  */
void
printf_mtslots_dbug_on(void)
{
  gint i;

  DBUG_ENTER("printf_mtslots_dbug_on");

  for (i=0;  i < MAX_SLOTS; i++) {
    if (mtmem[i].address != NULL) {
      printf("slot:%5i  a:%p  s:%5i  f:%32s  l:%4i\n", i, 
             mtmem[i].address, mtmem[i].size, mtmem[i].file, mtmem[i].line);
    } //if
  } //for i

  DBUG_VOID_RETURN;
} //printf_mtslots_dbug_on


/** MT_STRing_DUPlication copies a given string into fresh allocated memory.
  * @return a pointer to the fresh allocated string or NULL if the 
  *         given char* is NULL.
  */
gchar* 
mtstrdup_dbug_on(gchar* src, gchar* file, gint line)
{
  gchar* dst = NULL;

  DBUG_ENTER("mtstrdup_dbug_on");

  if (src != NULL) {
    dst = mtmalloc_dbug_on(strlen(src) + 1, file, line);
    strcpy(dst, src);
  }

  DBUG_RETURN(dst);
} //mtstrdup_dbug_on


gchar*
mtstrdup_dbug_off(gchar* src)
{
  gchar* dst = NULL;

  DBUG_ENTER("mtstrdup_dbug_off");

  if (src != NULL) {
    dst = g_malloc0(strlen(src) + 1);
    strcpy(dst, src);
  }

  DBUG_RETURN(dst);
} //mtstrdup_dbug_off


/* Concatenats a number of strings. If one of the strings is NULL then it 
 * will be ignorerd, but at least one string should be unequal to NULL!
 * @param nrArgs number of strings to be concatenated
 * @param ... all strings to be concatenated
 * @return the concatenated string in fresh (and exact) allocated memory */
gchar* 
mtstrconcat(int nrArgs, ...)
{
  gint i, len;
  gchar *tmpStr, *res;
  va_list ap; /* ap points to each argument */

  DBUG_ENTER("mtstrconcat");
  /* get length of all strings */
  va_start(ap, nrArgs); /* initialization for va_arg */
  len = 0;
  i   = 1;
  while (i <= nrArgs) {
    tmpStr = va_arg(ap, gchar*); /* get actual argument */
    if (tmpStr !=  NULL) {
      len += strlen(tmpStr);
    }
    i++;
  }
  va_end(ap); /* finalization for va_arg */


  res = (gchar*) mtmalloc(len+1);


  /* concatenate strings */
  va_start(ap, nrArgs); /* initialization for va_arg */
  i = 1;
  if (i <= nrArgs) {
    tmpStr = va_arg(ap, gchar*); /* get actual argument */
    if (tmpStr !=  NULL) {
      strcpy(res, tmpStr);
    } else {
      strcpy(res, "");
    }
    i++;
  }
  while (i <= nrArgs) {
    tmpStr = va_arg(ap, gchar*); /* get actual argument */
    if (tmpStr !=  NULL) {
      strcat(res, tmpStr);
    }
    i++;
  }
  va_end(ap); /* finalization for va_arg */

  DBUG_RETURN(res);
} //mtstrconcat


