#ifndef _MY_MAIN_H
#define _MY_MAIN_H


#include "my_dbug.h"


/******************************************************************************
 *
 * This file defines some (ugly but usefull) macros for handling command line 
 * switches, which are provided by the two arguments of the main-function 
 * (int argc and char* argv[]).
 * At first some examples to get an idea of how these macros are able to serve
 * you.
 *
 * The essential parts are:
 * ------------------------
 * int MAIN {
 *   BEGIN_ARGUMENTS
 *     LONG_OPTIONS
 *     SHORT_OPTIONS
 *     NOT_A_OPTION {... _actarg ...}
 *   END_ARGUMENTS
 *
 *   ...
 *   return 0;
 * }
 *
 * Using dbug-options only (see dbug-package of Fred Fish)
 * -------------------------------------------------------
 * int MAIN {
 *   BEGIN_ARGUMENTS
 *     DBUG_OPTIONS
 *     LONG_OPTIONS
 *     SHORT_OPTIONS
 *     NOT_A_OPTION {... _actarg ...}
 *   END_ARGUMENTS
 *
 *   ...
 *   return 0;
 * }
 *
 * An example with long, short and dbug options
 * --------------------------------------------
 * int MAIN {
 *   BEGIN_ARGUMENTS
 *     DBUG_OPTIONS
 *     LONG_OPTIONS
 *       LOPT("help") {...}
 *       LOPT("file") PARAM {... _param ...}
 *       OTHER_LOPT {... _actarg ...}
 *     SHORT_OPTIONS
 *       SOPT('?') {...}
 *       SOPT('f') PARAM {... _param ...}
 *       OTHER_SOPT {... _actarg ...}
 *     NOT_A_OPTION {... _actarg ...}
 *   END_ARGUMENTS
 *
 *   ...
 *   return 0;
 * }
 *
 * The ...-parts may be empty of filled with your own code. The variables
 * _argc, _actarg and _param are only valid within BEGIN_ARGUMENTS and
 * END_ARGUMENTS. The int _argc provides the number of the argument actually 
 * scanned for command line switches, char* _actarg provides the actual
 * argument and char* _param provides an additional parameter of a command 
 * line switch, if PARAM is used befor.
 * At the command line long options are introduced by "--", dbug options
 * by "-#" and short options by "-", where each following char represents one
 * command line switch. Everything else is NOT_A_OPTION.
 * OTHER_xOPT are optional. But if they are used, thay have to be the last 
 * macro within the corresponding list.
 *
 * Now an example of concrete code and its output:

#include <stdio.h>
#include "my_main.h"

int MAIN
{
  BEGIN_ARGUMENTS
    DBUG_OPTIONS
    LONG_OPTIONS
      LOPT("file") PARAM { printf("LOPT: file with parameter '%s'\n", _param);
                         }
      LOPT("help") { printf("LOPT: help\n"); }
      OTHER_LOPT   { printf("OTHER_LOPT '%s'\n", _actarg); }
    SHORT_OPTIONS
      SOPT('f')    PARAM { printf("SOPT: f with parameter '%s'\n", _param); }
      SOPT('d')    PARAM { printf("SOPT: d with parameter '%s'\n", _param); }
      SOPT('s')    { printf("SOPT: s\n"); }
      OTHER_SOPT   { printf("OTHER_SOPT '%.1s'\n", _actarg); }
    NOT_A_OPTION { printf("NOT_A_OPTION '%s'\n", _actarg); }
  END_ARGUMENTS

  return 0;
}
 * 
> a.out -#d -sfda f1 f2 --file f3  -? --help --long f4 
SOPT: s
SOPT: f with parameter 'f1'
SOPT: d with parameter 'f2'
OTHER_SOPT 'a'
LOPT: file with parameter 'f3'
OTHER_SOPT '?'
LOPT: help
OTHER_LOPT 'long'
NOT_A_OPTION 'f4'  
>
 *
 * The dbug options are not printed out, they will be delivered to the 
 * dbug-package by using DBUG_PUSH(string); (string = "d" in the case above).
 *
 *****************************************************************************/



#define MAIN                                                                  \
    main(int argc, char *argv[])

#define BEGIN_ARGUMENTS                                                       \
  {                                                                           \
    int  _argc = 1, nr_params = 0;                                            \
    char *_actarg = NULL, *_param = NULL;                                     \
    while (_argc < argc) {                                                    \
      if (argv[_argc][0] == '-') {                                            \
        /* ISOPTION */                                                        \
        switch (argv[_argc][1]) {                                             \
        case   0: break;

#define DBUG_OPTIONS                                                          \
                  break;                                                      \
	case '#':  /* DBUGOPTIONS */                                          \
                  _actarg = &(argv[_argc][2]);                                \
                  DBUG_PUSH(_actarg);

#define LONG_OPTIONS                                                          \
                  break;                                                      \
        case '-': /* LONGOPTIONS */                                           \
                  _actarg = &(argv[_argc][2]);                                \
                  if (0) {

#define LOPT(string)                                                          \
                    _param = NULL;                                            \
		  }                                                           \
                  else if (0 == strcmp(_actarg, string)) {

#define PARAM                                                                 \
                    nr_params += 1;                                           \
                    _param = argv[_argc + nr_params]; 

#define OTHER_LOPT                                                            \
                    _param = NULL;                                            \
		  }                                                           \
                  else {

#define SHORT_OPTIONS                                                         \
                    _param = NULL;                                            \
		  }                                                           \
                  break;                                                      \
	default : /* SHORTOPTIONS */                                          \
                  _actarg = &(argv[_argc][1]);                                \
                  while(*_actarg != '\0') {                                   \
                    switch (*_actarg) {                                       \
                    case   0: break;

#define SOPT(character)                                                       \
                              _param = NULL; break;                           \
                    case character: 

#define OTHER_SOPT                                                            \
                              _param = NULL; break;                           \
                    default: 

#define NOT_A_OPTION                                                          \
                              _param = NULL; break;                           \
		    } /* switch */                                            \
                    _actarg++;                                                \
                  } /* while */                                               \
       	} /*switch */                                                         \
      }else {                                                                 \
        /* NOTAOPTION */                                                      \
        _actarg = argv[_argc]; 

#define END_ARGUMENTS                                                         \
      } /* if-else */                                                         \
      _argc += 1 + nr_params;                                                 \
      nr_params = 0;                                                          \
    } /* while */                                                             \
  }



#endif //_MY_MAIN_H
