/* %option noyywrap / now there exists a yywrap-function */

%{
/*
 * A lexer to scan a KIEL-ML program file.
 *
 * Differences to the old lexer:
 * - filters (even nested) comments
 * - accept  1.2e~3  as real instead of  1.2e-3
 * - correct definition of (alphanumeric) identifiers (see [Ull] chapter 3)
 * - less restrictions to strings; accept all signs within the quotation
 *   marks but \n and a quotation mark itself (\")
 *   (see [Ull] chapter 1)
 * - return "int list" as two token instead of one, because "my_type list" is 
 *   a valid type now
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include "../Misc/my_dbug.h"
#include "../Misc/my_main.h"
#include "../Misc/utilities.h"
#include "../Misc/error_interface.h"
#include "../Misc/mem_management.h"
#include "../Misc/garbage_collection.h"
#include "../Evaluator/ml_limits.h"
#include "resources.h"
#include "syntaxtree.h"
#include "expr.h"
#include "expr_parser.h"

/* A macro to set the last recogniced token as a string and return the 
 * corresponding int value. In xyz_RETURN(FUN) #arg will be expanded to "FUN".
 * Using "arg" instead of #arg would not be expanded. It would still be "arg".
 */
#define NOVAL_RETURN(arg) { yylval.noval.lineno = yyylineno;                  \
                            yylval.noval.colno  = yyycolno - yyleng;          \
                            lastTokenStr = #arg;                              \
                            return(arg);                                      \
                          }
#define LEX_RETURN(arg) { lastTokenStr = #arg;                                \
                          return(arg);                                        \
                        }




static int yyylineno = 1; //some lex-programms use an own yylineno
static int yyycolno  = 1;
static int comment_level = 0;
static char* lastTokenStr;

//forward declarations
static void setYylval_mlval_mlstring(char* srcstr);
void setYylval_strval(struct _Strval_struct *dstval, 
                      char* srcstr, int line, int col);

#undef YY_INPUT
#define YY_INPUT(buf, res, max_size) (res = str_exprinput(buf, max_size))
char* expr_input    = NULL;
char* expr_inputptr = NULL;
char* expr_inputlim = NULL;

int
str_exprinput(char* buf, int max_size)
{
  int rest_size = (int)(expr_inputlim - expr_inputptr);
  int n         = (max_size < rest_size) ? max_size : rest_size;
  
  if (n > 0) {
    memcpy(buf, expr_inputptr, n);
    expr_inputptr += n;
  }
  return n;
}
%}

%x XSS_COMMENT

ws                   [ \t]
digit                [0-9]
real_number          {digit}+\.{digit}+([eE][~+]?{digit}+)?
  /* for the definition of an (alphanumeric) identifier see [Ull] chapter 3 */
identifier           [a-zA-Z][a-zA-Z0-9_]*


%%


  /**************************************************************************
   * Some rules to filter (even nested) comments.
   **************************************************************************/
"(*"                 { yyycolno += 2;
                       comment_level++;
                       /* Switch to the eXclusiv Start State XSS_COMMENT,
                        * where ONLY those rules match, which start with
                        * <XSS_COMMENT>. */
                       BEGIN XSS_COMMENT; 
                     }
<XSS_COMMENT>"(*"    { yyycolno += 2;
                       comment_level++; /* for nested comments */
                     }
<XSS_COMMENT>"*)"    { yyycolno += 2;
                       comment_level--; 
                       if (comment_level == 0) { 
                         BEGIN 0; /* return to default start state */
                       }
                     }
<XSS_COMMENT>.       { yyycolno++;
                     }
<XSS_COMMENT>\n      { yyycolno = 1;
                       yyylineno++; 
                     }



  /**************************************************************************
   * whitespaces, single signs and brackets
   **************************************************************************/

{ws}+                { yyycolno += yyleng; /* ignore white spaces */ }
\n                   { yyycolno = 1; yyylineno++; }
","                  { yyycolno += yyleng; NOVAL_RETURN(COMMA); }
":"                  { yyycolno += yyleng; NOVAL_RETURN(COLON); }
";"                  { yyycolno += yyleng; NOVAL_RETURN(SEMICOLON); }
"("                  { yyycolno += yyleng; NOVAL_RETURN(BRA_L); }
")"                  { yyycolno += yyleng; NOVAL_RETURN(BRA_R); }
"["                  { yyycolno += yyleng; NOVAL_RETURN(SQRBRA_L); }
"]"                  { yyycolno += yyleng; NOVAL_RETURN(SQRBRA_R); }
"|"                  { yyycolno += yyleng; NOVAL_RETURN(VBAR); }
"_"                  { yyycolno += yyleng; NOVAL_RETURN(UNDERSCORE); }


  /**************************************************************************
   *  keywords, constants and operators
   **************************************************************************/

  /**** other keywords ****/
"datatype"           { yyycolno += yyleng; NOVAL_RETURN(DATATYPE); }
"of"                 { yyycolno += yyleng; NOVAL_RETURN(OF); }
"fun"                { yyycolno += yyleng; NOVAL_RETURN(FUN); }
"and"                { yyycolno += yyleng; NOVAL_RETURN(AND); }
"let val"            { yyycolno += yyleng; NOVAL_RETURN(LET); }
"in"                 { yyycolno += yyleng; NOVAL_RETURN(IN); }
"end"                { yyycolno += yyleng; NOVAL_RETURN(END); }
"if"                 { yyycolno += yyleng; NOVAL_RETURN(IF); }
"then"               { yyycolno += yyleng; NOVAL_RETURN(THEN); }
"else"               { yyycolno += yyleng; NOVAL_RETURN(ELSE); }


  /**** basic types ****/
"unit"               { yyycolno += yyleng; NOVAL_RETURN(UNIT); }
"bool"               { yyycolno += yyleng; NOVAL_RETURN(BOOL); }
"int"                { yyycolno += yyleng; NOVAL_RETURN(INT); }
"real"               { /* this rule matches the function real: int -> real
                        * as well as the type real */
                       yyycolno += yyleng; NOVAL_RETURN(REAL); }
"string"             { yyycolno += yyleng; NOVAL_RETURN(STRING); }
"list"               { yyycolno += yyleng; NOVAL_RETURN(LIST); }


  /**** constants ****/
  /* do not return '()' by LEX_RETURN(UNIT_CONST). '()' may belong to a 
   * parameterless function 'fun f(): int = 5;'.
   */

"true"               { yyycolno += yyleng;
                       yylval.mlval.val.mlbool = TRUE;
                       yylval.mlval.lineno     = yyylineno;
                       yylval.mlval.colno      = yyycolno - yyleng;
                       LEX_RETURN(BOOL_CONST);
                     }
"false"              { yyycolno += yyleng;
                       yylval.mlval.val.mlbool = FALSE;
                       yylval.mlval.lineno     = yyylineno;
                       yylval.mlval.colno      = yyycolno - yyleng;
                       LEX_RETURN(BOOL_CONST);
                     }
{digit}+             { gdouble tmp;
                       tmp = atof(yytext);
                       if (isIntBetweenLimits(tmp)) {
                         yyycolno += yyleng;
                         yylval.mlval.val.mlint  = (gint)tmp;
                         yylval.mlval.lineno     = yyylineno;
                         yylval.mlval.colno      = yyycolno - yyleng;
                         LEX_RETURN(INT_CONST);
                       } else {
                         LEX_RETURN(SCANNER_ERROR);
                       }
                     }
{real_number}        { gdouble tmp;
                       strreplacechar(yytext, '~', '-');
                       tmp = atof(yytext);
                       if (isRealBetweenLimits(tmp)) {
                         yyycolno += yyleng;
                         yylval.mlval.val.mlreal = tmp;
                         yylval.mlval.lineno     = yyylineno;
                         yylval.mlval.colno      = yyycolno - yyleng;
                         LEX_RETURN(REAL_CONST);
                       } else {
                         LEX_RETURN(SCANNER_ERROR);
                       }
                     }
  /* a string must not span more than one line and must not support     */
  /* escape sequences in KIEL-ML (standard ML supports escape sequences */
  /* see [Ull] chapter 1)                                               */
\"[^\"\n]*\"         { yyycolno += yyleng; 
                       setYylval_mlval_mlstring(yytext);
                       yylval.mlval.lineno = yyylineno;
                       yylval.mlval.colno  = yyycolno - yyleng;
                       LEX_RETURN(STRING_CONST);
                     }
"nil"                { yyycolno += yyleng;
                       yylval.mlval.val.mllist = NULL;
                       yylval.mlval.lineno     = yyylineno;
                       yylval.mlval.colno      = yyycolno - yyleng;
                       LEX_RETURN(LIST_CONST);
                     }


  /**** unary operators ****/
"not"                { yyycolno += yyleng; NOVAL_RETURN(UNARY_NOT); }
"~"                  { yyycolno += yyleng; NOVAL_RETURN(UNARY_MINUS); }


  /**** infix operators ****/
"andalso"            { yyycolno += yyleng; NOVAL_RETURN(ANDALSO); }
"orelse"             { yyycolno += yyleng; NOVAL_RETURN(ORELSE); }
"="                  { yyycolno += yyleng; NOVAL_RETURN(EQUAL); }
"<>"                 { yyycolno += yyleng; NOVAL_RETURN(UNEQUAL); }
"<"                  { yyycolno += yyleng; NOVAL_RETURN(LT); }
"<="                 { yyycolno += yyleng; NOVAL_RETURN(LE); }
">"                  { yyycolno += yyleng; NOVAL_RETURN(GT); }
">="                 { yyycolno += yyleng; NOVAL_RETURN(GE); }
"+"                  { yyycolno += yyleng; NOVAL_RETURN(PLUS); }
"-"                  { yyycolno += yyleng; NOVAL_RETURN(MINUS); }
"*"                  { yyycolno += yyleng; NOVAL_RETURN(MULT); }
"mod"                { yyycolno += yyleng; NOVAL_RETURN(MOD); }
"div"                { yyycolno += yyleng; NOVAL_RETURN(INT_DIV); }
"/"                  { yyycolno += yyleng; NOVAL_RETURN(REAL_DIV); }
"^"                  { yyycolno += yyleng; NOVAL_RETURN(STRING_CONCAT); }
"::"                 { yyycolno += yyleng; NOVAL_RETURN(PREPEND); }
"@"                  { yyycolno += yyleng; NOVAL_RETURN(LIST_CONCAT); }


  /**** functions ****/
"abs"                { yyycolno += yyleng; NOVAL_RETURN(ABS_); }
"min"                { yyycolno += yyleng; NOVAL_RETURN(MIN_); }
"max"                { yyycolno += yyleng; NOVAL_RETURN(MAX_); }
"sqrt"               { yyycolno += yyleng; NOVAL_RETURN(SQRT); }
"sin"                { yyycolno += yyleng; NOVAL_RETURN(SIN); }
"cos"                { yyycolno += yyleng; NOVAL_RETURN(COS); }
"exp"                { yyycolno += yyleng; NOVAL_RETURN(EXP); }
"ln"                 { yyycolno += yyleng; NOVAL_RETURN(LN); }
  /* the function "real" uses the same keyword as the type "real" */
"truncate"           { yyycolno += yyleng; NOVAL_RETURN(TRUNCATE); }
"size"               { yyycolno += yyleng; NOVAL_RETURN(SIZE); }
"ord"                { yyycolno += yyleng; NOVAL_RETURN(ORD); }
"chr"                { yyycolno += yyleng; NOVAL_RETURN(CHR); }
"substring"          { yyycolno += yyleng; NOVAL_RETURN(SUBSTRING); }
"null"               { yyycolno += yyleng; NOVAL_RETURN(NULL_); }
"hd"                 { yyycolno += yyleng; NOVAL_RETURN(HD); }
"tl"                 { yyycolno += yyleng; NOVAL_RETURN(TL); }
"length"             { yyycolno += yyleng; NOVAL_RETURN(LENGTH); }
"project"             { yyycolno += yyleng; NOVAL_RETURN(PROJECT); }
"rev"                { yyycolno += yyleng; NOVAL_RETURN(REV); }


  /**************************************************************************
   * identifiers, errors and misc
   **************************************************************************/

{identifier}         { yyycolno += yyleng;
                       setYylval_strval(&(yylval.strval),
                                        yytext, yyylineno, yyycolno - yyleng);
                       LEX_RETURN(ID);
                     }

\"[^\"\n]*           { /* error: unclosed string */
                       yyerror("unclosed string");
                       LEX_RETURN(SCANNER_ERROR);
                     }

.                    { yyycolno += yyleng;
                       ECHO;
                     }


%%


/** Dynamicly allocated memory used for the yylval will never be freed again
  * (e.g.: yyval.void_ptr = malloc(...)). So i copy the pointer to allocated
  * memory to a garbage collection and free them all at a time, when i know 
  * they are not needed anymore (thats at least when the yyparse() returns).
  */
void setYylval_strval(struct _Strval_struct *dstval, 
                      char* srcstr, int line, int col)
{
  DBUG_ENTER("setYylval_strval");

  dstval->str    = mtstrdup(srcstr);
  gcAddItem(dstval->str);
  dstval->lineno = line;
  dstval->colno  = col;

  DBUG_VOID_RETURN;
} //setYylval_strval


static void setYylval_mlval_mlstring(char* srcstr)
{
  DBUG_ENTER("setYylval_mlval_mlstring");

  srcstr[strlen(srcstr)-1] = '\0';
  yylval.mlval.val.mlstring = mtstrdup(srcstr+1);
  //gcAddItem(yylval.mlval.val.mlstring);

  DBUG_VOID_RETURN;
} //setYylval_mlval_mlstring


/** Should be called from yyy_Initialize only
  */
void yyy_InitScanner(void)
{
  DBUG_ENTER("yyy_InitScanner");

  yyylineno     = 1;
  yyycolno      = 1;
  comment_level = 0;

  DBUG_VOID_RETURN;
}


void
yyyerror_cleanup(void)
{
  DBUG_ENTER("yyyerror_cleanup");

  //clean up some things (e.g. free mem)
  YY_FLUSH_BUFFER;
  tree_parts        = free_treeparts(tree_parts);
  undefined_symbols = free_symboltable(undefined_symbols);
  binding_table     = free_symboltable(binding_table);

  DBUG_VOID_RETURN;
} //yyyerror_cleanup


int yyerror(char* msg)
{
  const int c_max = 48;
  char line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
       line5[c_max];

  DBUG_ENTER("yyerror");

  DBUG_PRINT("", ("l:%.3d c:%.2d '%s' at '%s'\n", 
                  yyylineno, yyycolno-yyleng, msg, yytext));

  //clean up some things (e.g. free mem)
  yyyerror_cleanup();

  //set error text
  sprintf(line0, TEXT_PROGRAMM_PARSE_FEHLER0, msg);
  sprintf(line1, TEXT_PROGRAMM_PARSE_FEHLER1, yyylineno);
  sprintf(line2, TEXT_PROGRAMM_PARSE_FEHLER2, yyycolno);
  sprintf(line3, TEXT_PROGRAMM_PARSE_FEHLER3, lastTokenStr);
  sprintf(line4, TEXT_PROGRAMM_PARSE_FEHLER4, yytext);
  sprintf(line5, TEXT_PROGRAMM_PARSE_FEHLER5);

  //check afterwards, if local variable lineX was big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  /* Avoid that an scanner-error will be overwritten by syntax-error.
   * The scanner returns the SCANNER_ERROR-token, which causes the parser
   * to detect a syntax-error.
   */
  if (isErrorDescriptionClear()) {
    set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                         line2, line3, line4, line5);
  }
  DBUG_RETURN(1);
} //yyerror


int yywrap(void)
{
  DBUG_ENTER("yywrap");
  DBUG_RETURN(1);
} //yywrap


/* for tests only
int MAIN
{
  DBUG_ENTER("main");

  BEGIN_ARGUMENTS
    DBUG_OPTIONS
    LONG_OPTIONS
    SHORT_OPTIONS
    NOT_A_OPTION {}
  END_ARGUMENTS

  while (0 != yylex()) {;}

  DBUG_RETURN(0);
}
a */

