#ifndef _SYNTAXTREE_H
#define _SYNTAXTREE_H



#include <gtk/gtk.h>
#include "symboltable.h"

#define EXPR_NODE_MAX_CHILDREN 3


/* The names of the nodetypes for operators and basic functions should be 
 * similar to the tokennames returned by the lexer. The Nodetype e_arg_list
 * represents the (not fixed in size) list of arguments for a user defined
 * function or constructor. One child points to the argument expression and 
 * another to the next e_arg_list (which points to the next argument) or
 * NULL if the list ends.
 */
typedef enum { e_invalid, e_errornode, 
	       e_param_list, e_arg_list, e_listkeeper, e_list, 
               e_constant, e_nil_const, e_variable, 
               e_unary_not, e_unary_minus, 
               e_andalso, e_orelse, e_equal, e_unequal, e_lt, e_le, e_gt, e_ge,
               e_plus, e_minus, e_mult, e_mod, e_int_div, e_real_div, 
               e_string_concat, e_prepend, e_list_concat,
               e_abs, e_min, e_max, e_sqrt, e_sin, e_cos, e_exp, e_ln, 
               e_real, e_truncate, e_size, e_ord, e_chr, e_substring,
               e_null, e_hd, e_tl, e_length, e_rev,
               e_ifthenelse,
               e_constructor, e_userfun, e_wildcard,
	       e_product,
	       e_letvars_list, e_let, e_project
} Nodetype;

typedef enum { 
  e_default, e_boxed, e_hidden
} Subnodetype;

typedef union {
  gpointer mlunit;
  gboolean mlbool;
  gint     mlint;
  gdouble  mlreal;
  gchar*   mlstring;
  GList*   mllist;
} MLValue;


/* MLType is the type of the value which is represented (should by filled by 
 * the typechecker at least). A enum is not possible, because of the 
 * user defined datatypes.
 *   0: not set yet; 1 - 5: unit, bool, int, real, string
 *  >5: user defined datatype
 * When e.g. 'int list' occurs and it has no type index yet , it gets a new 
 * type index like any userdefined type.
 */
typedef gint MLType;


typedef struct _Expr_node_struct{
  Nodetype   nodetype;  //the kind of this node
  Subnodetype subnodetype;
  gchar*     strrep;    //string representation of node
  MLValue    value;     //the value represented by this node 
                        //(may by unused)
  gchar*     dummy_strval; //to free a string value (value.mlstring)
  MLType     valuetype; //type of the value found/computed by the typechecker
  MLType     valuetype_expected; //expected type of the value
  gpointer   misc;      //additional stuff for later use 
                        //(e.g. the position on screen)
  //Maybe the following data should be moved into a misc-structure.
  //So use ACCESSMACROS!!!
  gint       lineno;    //line where the token was read
  gint       colno;     //col where the token was read
  gint       expansiondepth; //of this node during evaluation
 
  struct _Expr_node_struct *child[EXPR_NODE_MAX_CHILDREN];
} Expr_node_struct, *Expr_node_ptr;


typedef struct _Fun_node_struct {
  gchar*    fun_id;
  Nodetype  nodetype;    //only e_invalid, e_userfun and 
                         //e_constructor are used
  MLType    resulttype;  //type of the value which is returned or 
                         //constructed
  GList*    paramslist;  //list of Expr_node_ptr's for each clause
  GList*    bindingslist;//list of GList*'s for each clause one symboltable
                         //stores the type of id's occurring in parameters
  GList*    bodylist;    //unused in case of e_constructor, otherwise a
                         //list of Expr_node_ptr's for each clause
} Fun_node_struct, *Fun_node_ptr;


/* Some macros to access the children of an expr_node */
#define ARGLIST_EXPR_NODE(ptr)       ((ptr)->child[0])
#define PARAMLIST_EXPR_NODE(ptr)     ((ptr)->child[0])
#define ARG_EXPR_NODE(ptr)           ((ptr)->child[0])
#define PARAM_EXPR_NODE(ptr)         ((ptr)->child[0])
#define NEXT_EXPR_NODE(ptr)          ((ptr)->child[1])
#define LIST_EXPR_NODE(ptr)          ((ptr)->child[0])
#define LISTITEM_EXPR_NODE(ptr)      ((ptr)->child[0])
#define LISTNEXT_EXPR_NODE(ptr)      ((ptr)->child[1])
#define OPERAND_EXPR_NODE(ptr)       ((ptr)->child[0])
#define LEFT_OPERAND_EXPR_NODE(ptr)  ((ptr)->child[0])
#define RIGHT_OPERAND_EXPR_NODE(ptr) ((ptr)->child[1])
#define COND_EXPR_NODE(ptr)          ((ptr)->child[0])
#define THEN_EXPR_EXPR_NODE(ptr)     ((ptr)->child[1])
#define ELSE_EXPR_EXPR_NODE(ptr)     ((ptr)->child[2])

#define LET_VARS_EXPR_NODE(ptr)          ((ptr)->child[0])
#define LET_VAL_EXPR_NODE(ptr)     ((ptr)->child[1])
#define LET_EXPR_EXPR_NODE(ptr)     ((ptr)->child[2])


/* Other access-macros */
#define LINENO_NODE(ptr)             ((ptr)->lineno)
#define COLNO_NODE(ptr)              ((ptr)->colno)


#include "yylval_types.h"


extern GList* syntax_tree;
extern GList* tree_parts;
extern gboolean keep_tree_parts_in_mind;

extern void (*fpFree_expr_node_misc) (gpointer pMisc);


extern GList*
myGlistRemove(GList* list, gpointer data);



extern Expr_node_ptr
create_dummy(void);

extern Expr_node_ptr
create_errornode(void);

extern Expr_node_ptr
create_expr_param_list(Expr_node_ptr arg, Expr_node_ptr next);

extern Expr_node_ptr
create_expr_letvars_list(Expr_node_ptr arg, Expr_node_ptr next);

extern Expr_node_ptr
create_expr_let(Expr_node_ptr vars, Expr_node_ptr val,Expr_node_ptr expr);

extern Expr_node_ptr
create_expr_project(Expr_node_ptr pos,     
                    Expr_node_ptr expr);
    
extern Expr_node_ptr
create_expr_arg_list(Expr_node_ptr arg, Expr_node_ptr next);

extern Expr_node_ptr
create_expr_listkeeper(struct _Mlval_struct value, Expr_node_ptr list);

extern Expr_node_ptr
create_expr_list(Expr_node_ptr elm, Expr_node_ptr next);

extern Expr_node_ptr
create_expr_product(Expr_node_ptr list);

extern Expr_node_ptr
create_expr_constant(struct _Mlval_struct value, MLType valuetype_found);

extern Expr_node_ptr
create_expr_nil_const(struct _Mlval_struct value, MLType valuetype_found);

extern Expr_node_ptr
create_expr_variable(struct _Strval_struct strval);

extern Expr_node_ptr
create_expr_specialtype_param(gchar* var_id, MLType valuetype_expected);

extern Expr_node_ptr
create_expr_producttype_param(gchar* var_id, MLType valuetype_expected);

extern Expr_node_ptr
create_expr_unary_not(struct _Noval_struct noval, Expr_node_ptr operand);

extern Expr_node_ptr
create_expr_unary_minus(struct _Noval_struct noval, Expr_node_ptr operand);

extern Expr_node_ptr
create_expr_infixop(struct _Noval_struct noval, Nodetype type, 
                    Expr_node_ptr left_operand, 
                    Expr_node_ptr right_operand);
extern Expr_node_ptr
create_expr_basicfun(struct _Noval_struct  noval, 
                     Nodetype type, Expr_node_ptr arg);

extern Expr_node_ptr
create_expr_basicfun2(struct _Noval_struct  noval, 
                      Nodetype type, Expr_node_ptr arg1, Expr_node_ptr arg2);

extern Expr_node_ptr
create_expr_substring(struct _Noval_struct noval, Expr_node_ptr str, 
                      Expr_node_ptr startpos, Expr_node_ptr length);
extern Expr_node_ptr
create_expr_ifthenelse(struct _Noval_struct noval, Expr_node_ptr cond,     
                       Expr_node_ptr then_expr, Expr_node_ptr else_expr);
extern Expr_node_ptr
create_expr_constructor(struct _Strval_struct strval, Expr_node_ptr arglist);

extern Expr_node_ptr
create_expr_userfun(struct _Strval_struct strval, Expr_node_ptr arglist);

extern Expr_node_ptr
create_expr_wildcard(struct _Noval_struct noval);

extern Expr_node_ptr
free_expr_node(Expr_node_ptr node);

extern gchar* 
nodetype2str(Nodetype nt);

extern void
printf_expr_node(Expr_node_ptr node);



extern Fun_node_ptr
create_fun_node(gchar* fun_id, Nodetype nodetype, Expr_node_ptr params,
                MLType resulttype, Expr_node_ptr body, 
                GList* bindings);
extern void
append_fun_node(Fun_node_ptr pFunnode, 
		Expr_node_ptr params, GList* bindings, Expr_node_ptr body);
extern Fun_node_ptr
free_fun_node(Fun_node_ptr node);



extern GList*
free_syntaxtree(GList* st);

extern void 
printf_treeparts(GList* tree_parts);

extern GList*
free_treeparts(GList* tree_parts);


extern void 
nav_expr_node(Expr_node_ptr node);

extern void 
nav_syntaxtree(GList* st);

extern gchar* 
funNode2Signature(Fun_node_ptr pFun);

extern gchar*
funNode2DatatypeStr(Fun_node_ptr pFun);

extern gchar* 
exprNode2String(Expr_node_ptr pExpr);

extern gint 
getIndexOfChild(Expr_node_ptr parent, Expr_node_ptr child);

extern void 
replaceSubtree(Expr_node_ptr parent_of_old, Expr_node_ptr old, 
	       Expr_node_ptr new);

extern Expr_node_ptr 
duplicateSingleExprnode(Expr_node_ptr src);

extern Expr_node_ptr 
duplicateSubtree(Expr_node_ptr src);

extern Expr_node_ptr 
duplicateSubtreeButSetNewExpansiondepth(Expr_node_ptr src, 
					gint new_expansiondepth);



#endif //_SYNTAXTREE_H
