#include <gtk/gtk.h>
#include <string.h>
#include "../Misc/my_dbug.h"
#include "../Misc/mem_management.h"
#include "../Syntaxtree/symboltable.h"
#include "../Syntaxtree/syntaxtree.h"
#include "tc_util.h"


/* forward declarations */
MLType checkBody(Expr_node_ptr node);



static MLType 
addNotExistingListtype(MLType listitem_type)
{
  gchar* type_str;
  gint new_type_index;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("addNotExistingListtype");

  new_type_index = g_list_length(symbol_table) + 1;
  pSte = seekSymbolTypeindex(symbol_table, listitem_type);
  DBUG_ASSERT(pSte != NULL);

  type_str = mtmalloc(sizeof(gchar) * (strlen(pSte->id) + 6));
  strcpy(type_str, pSte->id);
  strcat(type_str, " list");

  symbol_table = addSymbol(symbol_table, type_str,  e_st_basictype,0, 
			   new_type_index, listitem_type, NULL);

  mtfree(type_str);

  DBUG_RETURN(new_type_index);
} //addNotExistingListtype


/** @return the type_index of this node, or 0 in case of an error.
  */
static MLType
checkBasicfun1Arg(MLType arg_type, MLType res_type, Expr_node_ptr node)
{
  MLType found_type, res;
  DBUG_ENTER("checkBasicfun1Arg");

  found_type = checkBody(OPERAND_EXPR_NODE(node));
  //printf("arg: %d   res: %d   found: %d\n", arg_type, res_type, found_type);
  if (found_type == 0) {
    res = 0;
  } else {
    if (!isFoundTypeCompatibleToExpected(arg_type, found_type)) {
      tc_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
	       COLNO_NODE( OPERAND_EXPR_NODE(node)),
	       arg_type, found_type);
      res = 0;
    } else {
      res = res_type;
    } //if
  }

  DBUG_RETURN(res);
} //checkBasicfun1Arg


/** @return the type_index of this node, or 0 in case of an error.
  */
static MLType
checkBasicfun2Args(MLType arg1_type, MLType arg2_type, 
		   MLType res_type, Expr_node_ptr node)
{
  MLType found_type1, found_type2, res;
  DBUG_ENTER("checkBasicfun2Args");

  found_type1 = checkBody(LEFT_OPERAND_EXPR_NODE(node));
  if (found_type1 == 0) { DBUG_RETURN(0); }
  found_type2 = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
  if (found_type2 == 0) { DBUG_RETURN(0); }

  if        ( !isFoundTypeCompatibleToExpected(arg1_type, found_type1) ) {
    tc_error(LINENO_NODE(LEFT_OPERAND_EXPR_NODE(node)), 
	     COLNO_NODE( LEFT_OPERAND_EXPR_NODE(node)),
	     arg1_type, found_type1);
    res = 0;
  } else if ( !isFoundTypeCompatibleToExpected(arg2_type, found_type2) ) {
    tc_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
	     COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
	     arg2_type, found_type2);
    res = 0;
  } else {
    res = res_type;
  }

  DBUG_RETURN(res);
} //checkBasicfun2Args



/** @return the type_index of this node, or 0 in case of an error.
  */
static MLType
checkUserfun(Expr_node_ptr node)
{
  gint                  actarg_count;
  MLType                found_type, res;
  Symboltable_entry_ptr pSte;
  Expr_node_ptr         pDefinedParams, pActArgs;
  Fun_node_ptr          pFNode;

  DBUG_ENTER("checkUserfun");

  //get the result_type
  pSte = seekSymbolId(symbol_table, node->strrep);
  DBUG_ASSERT(pSte != NULL);
  res = pSte->type_index;

  pActArgs = ARGLIST_EXPR_NODE(node);
  //pActArgs may be NULL in case of a parameterless constructor

  pFNode = pSte->definition_point;
  DBUG_ASSERT(pFNode != NULL);
  if (pFNode->paramslist == NULL) {
    actarg_count = 0;
    //count actParams
    while (pActArgs != NULL) {
      DBUG_ASSERT(pActArgs->nodetype == e_arg_list);
      actarg_count += 1;
      pActArgs = NEXT_EXPR_NODE(pActArgs);
    }
    if (actarg_count != 0) {
      DBUG_ASSERT(SYMBOL_NR_PARAMS(pSte) == 0);
      pc_error(LINENO_NODE(node), COLNO_NODE(node), 
	       SYMBOL_NR_PARAMS(pSte), actarg_count);
      res = 0;
    }
  } else {
    //the param-types are equal in each clause!!! So check the 1st only.
    pDefinedParams = pFNode->paramslist->data;
    actarg_count = 0;
    while ( (pDefinedParams != NULL) && (pActArgs != NULL)
	    && (res != 0) ) {
      DBUG_ASSERT(pActArgs->nodetype == e_arg_list);
      actarg_count += 1;
      found_type = checkBody(ARG_EXPR_NODE(pActArgs));
      if (found_type == 0) { DBUG_RETURN(0); }
      if ( !isFoundTypeCompatibleToExpected(
			       ARG_EXPR_NODE(pDefinedParams)->valuetype, 
			       found_type) ) {
	tc_error(LINENO_NODE(ARG_EXPR_NODE(pActArgs)), 
		 COLNO_NODE( ARG_EXPR_NODE(pActArgs)),
		 ARG_EXPR_NODE(pDefinedParams)->valuetype, found_type);
	res = 0;
      }

      pDefinedParams = NEXT_EXPR_NODE(pDefinedParams);
      pActArgs       = NEXT_EXPR_NODE(pActArgs);
    } //while

    /* check if defined == found number of parameters */
    if ( ((pDefinedParams != NULL) || (pActArgs != NULL))
	 && (res != 0) ) {
      //count the rest of the actArgs if any
      while (pActArgs != NULL) {
	actarg_count += 1;
	pActArgs = NEXT_EXPR_NODE(pActArgs);
      } //while
      pc_error(LINENO_NODE(node), COLNO_NODE(node), 
	       SYMBOL_NR_PARAMS(pSte), actarg_count);
      res = 0;
    } //if

  } //if pFNode->paramlist == NULL

  DBUG_RETURN(res);
} //checkUserfun

/** Sets the MLType for all occurences of a given Variable
  */
void
setMLTypeAllOccurences(gchar* ofVar, 
			MLType mltype,
                        Expr_node_ptr inExpr)
{
  gint i;

  DBUG_ENTER("setMLTypeAllOccurences");
  DBUG_ASSERT(ofVar            != NULL);
  DBUG_ASSERT(inExpr           != NULL);

g_print("NEWSET ");

  if ( (inExpr->nodetype == e_variable) && (inExpr->strrep != NULL)
       && (0 == strcmp(ofVar, inExpr->strrep)) )  {
	inExpr->valuetype = mltype;
  } else {
    
    // visibility for let statements
    if ( inExpr->nodetype == e_let ) {
      // replace all occurences in the expression-part only if the 
      // variable is not defined in the let 
      Expr_node_ptr vars = LET_VARS_EXPR_NODE(inExpr);
      int variable_is_defined_here = 0;
      while ( vars != NULL ) {
        if ( 0 == strcmp(ARG_EXPR_NODE(vars)->strrep,ofVar)  ) {
             variable_is_defined_here = 1;
             vars = NULL;
        } else {
          vars = NEXT_EXPR_NODE(vars);
        }
      }
      if ( ! variable_is_defined_here ) {
	setMLTypeAllOccurences(ofVar, mltype, LET_EXPR_EXPR_NODE(inExpr));
      }
      // always replace all occurences in the value-part
      setMLTypeAllOccurences(ofVar, mltype,LET_VAL_EXPR_NODE(inExpr));
      // never replace the variable-definition-part
    } else {
      
      for (i=0;  i < EXPR_NODE_MAX_CHILDREN;  i++) {
        if (inExpr->child[i] != NULL) {
	     	setMLTypeAllOccurences(ofVar, mltype,  inExpr->child[i]);
        } //for i
      }
    
    } //if == e_let
  }
g_print("...done\n");
  DBUG_VOID_RETURN;
} //setMLTypeAllOccurences


/** Check the body in a bottom-up analysis. For the body
  * both the root-node of this subtree and all leafs are allready typed, but 
  * operators like EQUAL and UNEQUAL make a top-down analysis difficult.
  * @param node is the actual node which has to be checked.
  * @return the type_index of this node, or 0 in case of an error.
  */
MLType
checkBody(Expr_node_ptr node)
{
  MLType tmp_type1, tmp_type2, tmp_type3, listitem_type, res = 0;
  Symboltable_entry_ptr tmp_pSte;

  DBUG_ENTER("checkBody");
  DBUG_ASSERT(node != NULL);
  if (node == NULL) {
    DBUG_RETURN(res);
  }
  
  switch (node->nodetype) {
  case e_listkeeper:
    if (LIST_EXPR_NODE(node) == NULL) {
      res = list_type_index;
    } else {
      res = checkBody(LIST_EXPR_NODE(node));
      DBUG_ASSERT(res != list_type_index); 
      //res may be 'list list" but not 'list'
    }
    DBUG_PRINT( "tc", ("e_listkeeper - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_list:
    tmp_type1 = checkBody(LISTITEM_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }

    //is it the last real element?
    if (LISTNEXT_EXPR_NODE(node) == NULL) {
      //it is the last real element
      res = getListtypeForItemtype(tmp_type1);
      if (res <= 0) {
	res = addNotExistingListtype(tmp_type1);
      }
    } else {
      //it is NOT the last real element
      tmp_type2 = checkBody(LISTNEXT_EXPR_NODE(node));
      if (tmp_type2 == 0) { DBUG_RETURN(0); }

      //is tmp_type2 really a MLList?
      listitem_type = getItemtypeOfList(tmp_type2);
      DBUG_ASSERT(listitem_type > -1);//-1 : type not found in symbol_table
      if (listitem_type == 0) {
	//tmp_type2 is not a MLList
	tc_error(LINENO_NODE(LISTNEXT_EXPR_NODE(node)), 
		 COLNO_NODE( LISTNEXT_EXPR_NODE(node)),
		 list_type_index, tmp_type2);
	res = 0;
      } else {
	//tmp_type2 is a MLList
	tmp_type3 = getCompatibleType(tmp_type1, listitem_type);
	if (tmp_type3 == 0) {
	  tc_error(LINENO_NODE(LISTITEM_EXPR_NODE(node)), 
		   COLNO_NODE( LISTITEM_EXPR_NODE(node)),
		   listitem_type, tmp_type1);
	  res = 0;
	} else {
	  res = tmp_type2; //return the list-type and not the item-type!
	} //if types are compatible
      } //if tmp_type2 is really a MLList
    } //if last real element
    break;

    /* the old way
  case e_list:
    tmp_type1 = checkBody(LISTITEM_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }

    //is it the last element/nil-element?
    if (LISTNEXT_EXPR_NODE(node) == NULL) {
      res = list_type_index;
    } else {
      tmp_type2 = checkBody(LISTNEXT_EXPR_NODE(node));
      if (tmp_type2 == 0) { DBUG_RETURN(0); }
      //is it the element befor the last/nil-element?
      if (tmp_type2 == list_type_index) {
	res = getListtypeForItemtype(tmp_type1);
	if (res <= 0) {
	  res = addNotExistingListtype(tmp_type1);
	}
	LISTNEXT_EXPR_NODE(node)->valuetype = res;//set listtype of nil-element
      } else {
	//is tmp_type2 really a MLList?
	listitem_type = getItemtypeOfList(tmp_type2);
	DBUG_ASSERT(listitem_type > -1);//-1 : type not found in symbol_table
	if (listitem_type == 0) {
	  //tmp_type2 is not a MLList
	  tc_error(LINENO_NODE(LISTNEXT_EXPR_NODE(node)), 
		   COLNO_NODE( LISTNEXT_EXPR_NODE(node)),
		   list_type_index, tmp_type2);
	  res = 0;
	} else {
	  //tmp_type2 is a MLList
 	  tmp_type3 = getCompatibleType(tmp_type1, listitem_type);
	  if (tmp_type3 == 0) {
	    tc_error(LINENO_NODE(LISTITEM_EXPR_NODE(node)), 
		     COLNO_NODE( LISTITEM_EXPR_NODE(node)),
		     listitem_type, tmp_type1);
	    res = 0;
	  } else {
	    res = tmp_type2; //return the list-type and not the item-type!
	  }
	} //is tmp_type2 a MLLIst?
      }
    }
    DBUG_PRINT( "tc", ("e_list - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;
    */

    case e_project:

      {
      MLType left_type = checkBody(LEFT_OPERAND_EXPR_NODE(node));
      MLType right_type = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
      
      res = 0;
      
      // linke seite muss int sein
      if ( left_type == int_type_index && LEFT_OPERAND_EXPR_NODE(node)->nodetype == e_constant ) {
        
        // schauen, ob das rechte ein produkttyp ist
        Symboltable_entry_ptr rSymb = seekSymbolTypeindex(symbol_table, right_type);
        int pos = LEFT_OPERAND_EXPR_NODE(node)->value.mlint;
          
        if ( rSymb->symbol_type != e_st_product && pos == 0 ) {
          res = right_type;
        }  else {
          Expr_node_ptr part = ARG_EXPR_NODE(RIGHT_OPERAND_EXPR_NODE(node));
          while ( part != NULL && pos > 0 ) {
              part = NEXT_EXPR_NODE(part);;
              pos--;
          }
          if ( pos == 0 && part != NULL) {
            res = checkBody(ARG_EXPR_NODE(part));
          }
        }
		

      }
      if ( res == 0 ) {
        project_error(node->lineno, node->colno);
      }

      }
      break;
    
    case e_let:
      {
      // set the type information for all occurences of the variables, 
      // that are defined here (as they occur in the expr part)
      Expr_node_ptr vars = LET_VARS_EXPR_NODE(node);
      MLType expr_type;
      MLType vals_type;
      Symboltable_entry_ptr valsSymb;
      while ( vars != NULL ) {
        setMLTypeAllOccurences(
          ARG_EXPR_NODE(vars)->strrep,
          ARG_EXPR_NODE(vars)->valuetype_expected, 
          LET_EXPR_EXPR_NODE(node)
        );
        vars = NEXT_EXPR_NODE(vars);
      }


      expr_type = checkBody(LET_EXPR_EXPR_NODE(node));
      vals_type = checkBody(LET_VAL_EXPR_NODE(node));
      valsSymb = seekSymbolTypeindex(symbol_table, vals_type);
      vars = LET_VARS_EXPR_NODE(node);
      
      if ( valsSymb->symbol_type != e_st_product ) {

        if (  !isFoundTypeCompatibleToExpected(ARG_EXPR_NODE(vars)->valuetype_expected, vals_type) ) {
          res = 0;
        } else {
          res = expr_type;
        }
		
      } else {

        // compare the types of the defined variables and of the expr part
        Product_part_ptr part = valsSymb->product_parts;
        vars = LET_VARS_EXPR_NODE(node);
        res = expr_type;
 
        while ( vars != NULL && part != NULL ) {
	  if ( ! isFoundTypeCompatibleToExpected(ARG_EXPR_NODE(vars)->valuetype_expected, part->type) ) {
            res = 0;
            difftypes_error(LINENO_NODE(ARG_EXPR_NODE(vars)), 
		      COLNO_NODE( ARG_EXPR_NODE(vars)),
		      ARG_EXPR_NODE(vars)->valuetype_expected, part->type);
            vars = NULL;
          } else {
            vars = NEXT_EXPR_NODE(vars);
            part = part->next;
          }
        }
	if ( vars != NULL && res != 0 ) {
          res = 0;
          let_pc_error(LINENO_NODE(ARG_EXPR_NODE(vars)), 
		      COLNO_NODE(ARG_EXPR_NODE(vars)));

        }
        if ( part != NULL && res != 0 ) {
          res = 0;
          let_pc_error(LINENO_NODE(LET_VAL_EXPR_NODE(node)), 
		      COLNO_NODE(LET_VAL_EXPR_NODE(node)));
        }
	
      }
      }

      break;
      
  case e_product:
   
    {
     Expr_node_ptr n;
    Product_part_ptr parts = NULL;
    int new_type_index = g_list_length(symbol_table) + 1;
    char* tmp_str = "int";
    char* id_of_product = mtmalloc(1000);
    Symboltable_entry_ptr pSte ;

    
    Expr_node_ptr product_type = create_expr_param_list( 
        create_expr_producttype_param(tmp_str, new_type_index), 
    NULL);
    
     printf("xpm: check product\n");
//    printf("%s",printf_expr_nodeLISTITEM_EXPR_NODE(node));
    printf("valuetype expected: %i", node->valuetype_expected);
    printf_expr_node(node);

    printf_expr_node(LISTITEM_EXPR_NODE(node));
    //printf_expr_node(LISTNEXT_EXPR_NODE(node));

    // get types for all product-nodes
    
    
    
    id_of_product[0] = '\0';
    n = LISTITEM_EXPR_NODE(node);
    
    while ( n != NULL ) {
      
      Symboltable_entry_ptr pSte ;
      MLType part_type = checkBody(LISTITEM_EXPR_NODE(n));

      printf("part %i\n", part_type);
      printf_expr_node(LISTITEM_EXPR_NODE(n));

      if ( parts == NULL ) {
	parts = createProduct_part(part_type);
      } else {
	appendProduct_part(parts,part_type);
      }
      
      pSte = seekSymbolTypeindex(symbol_table, part_type);
      
      new_type_index++;
      
      printf(pSte->id);
      printf("\n"); 
      
      if ( strlen(id_of_product) > 0 ) {
        strncat(id_of_product, " * ", 3);
      }
      strncat(id_of_product, pSte->id, 50);

      
      n = LISTNEXT_EXPR_NODE(n);
    }

    printf("complete type is %s\n",id_of_product);
    product_type  = 0;
    
    pSte = seekSymbolId(symbol_table, id_of_product);

    if ( pSte == 0) {  
      int new_type_index = g_list_length(symbol_table) + 1;
      symbol_table = addSymbol(symbol_table, 
                             id_of_product, e_st_product, 0,
                             new_type_index, 
                             0, NULL);
      pSte = seekSymbolId(symbol_table, id_of_product);
      pSte->product_parts = parts;
    }
  
    if ( pSte == 0) {  
      printf("couldn't find symb!!! %i\n\n", 0); 
    }
    
    printf("xxx\n"); 
    printf("found symb %i", pSte->type_index); 

    res = pSte->type_index;

    printf("result-type is: %i\n", res);
    //printf("%i\n", checkBody(LISTNEXT_EXPR_NODE(node)));
    printf("xxx\n");
   // res = checkBody(LISTITEM_EXPR_NODE(node));

    }
    break;

  case e_constant:
    res = node->valuetype;
    //later on node->valuetype will be set to res, but thats ok
    DBUG_PRINT( "tc", ("e_constant - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_nil_const:
    res = node->valuetype;
    //later on node->valuetype will be set to res, but thats ok
    DBUG_PRINT( "tc", ("e_nil_const - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_variable:
    if ( node->valuetype > 0 ) {
      res = node->valuetype ;
    } else {
    tmp_pSte = seekSymbolId(act_binding, node->strrep);
    DBUG_ASSERT(tmp_pSte != NULL);
    res = tmp_pSte->type_index;
    DBUG_PRINT( "tc", ("e_variable - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    }
    break;

  case e_unary_not:
    res = checkBasicfun1Arg(bool_type_index, bool_type_index, node); 
    DBUG_PRINT( "tc", ("e_unary_not - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_unary_minus:
    tmp_type1 = checkBody(OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    if ( ( !isFoundTypeCompatibleToExpected(int_type_index,  tmp_type1) ) &&
         ( !isFoundTypeCompatibleToExpected(real_type_index, tmp_type1) ) ) {
      expectedX_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( OPERAND_EXPR_NODE(node)),
		      "int, real", tmp_type1);
      res = 0;
    } else {
      res = tmp_type1;
    }
    DBUG_PRINT( "tc", ("e_unary_minus - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_andalso:
  case e_orelse:
    res = checkBasicfun2Args(bool_type_index, bool_type_index, 
                             bool_type_index, node);
    DBUG_PRINT( "tc", ("e_andalso+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_equal:
  case e_unequal:
    tmp_type1 = checkBody(LEFT_OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }

    tmp_type3 = getCompatibleType(tmp_type1, tmp_type2);
    if (tmp_type3 == 0) {
      difftypes_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		      tmp_type1, tmp_type2);
      res = 0;
    } else {
      res = bool_type_index;
    }
    DBUG_PRINT( "tc", ("e_equal+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_lt:
  case e_le:
  case e_gt:
  case e_ge:
    tmp_type1 = checkBody(LEFT_OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }
    if        ( (!isFoundTypeCompatibleToExpected(int_type_index,    tmp_type1)) && 
		(!isFoundTypeCompatibleToExpected(real_type_index,   tmp_type1)) && 
		(!isFoundTypeCompatibleToExpected(string_type_index, tmp_type1))   ) {
      expectedX_error(LINENO_NODE(LEFT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( LEFT_OPERAND_EXPR_NODE(node)),
		      "int, real, string", tmp_type1);
      res = 0;
    } else if ( (!isFoundTypeCompatibleToExpected(int_type_index,    tmp_type2)) && 
		(!isFoundTypeCompatibleToExpected(real_type_index,   tmp_type2)) && 
		(!isFoundTypeCompatibleToExpected(string_type_index, tmp_type2))   ) {
      expectedX_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		      "int, real, string", tmp_type2);
      res = 0;
    } else if (tmp_type1 != tmp_type2) {
      difftypes_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		      tmp_type1, tmp_type2);
      res = 0;
    } else {
      res = bool_type_index;
    }
    DBUG_PRINT( "tc", ("e_lt+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_plus:
  case e_minus:
  case e_mult:
    tmp_type1 = checkBody(LEFT_OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }
    if        ( (!isFoundTypeCompatibleToExpected(int_type_index,  tmp_type1)) && 
	        (!isFoundTypeCompatibleToExpected(real_type_index, tmp_type1))   ) {
      expectedX_error(LINENO_NODE(LEFT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( LEFT_OPERAND_EXPR_NODE(node)),
		      "int, real", tmp_type1);
      res = 0;
    } else if ( (!isFoundTypeCompatibleToExpected(int_type_index,  tmp_type2)) && 
		(!isFoundTypeCompatibleToExpected(real_type_index, tmp_type2))   ) {
      expectedX_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		      "int, real", tmp_type2);
      res = 0;
    } else if (tmp_type1 != tmp_type2) {
      difftypes_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		      tmp_type1, tmp_type2);
      res = 0;
    } else {
      res = tmp_type1;
    }
    DBUG_PRINT( "tc", ("e_plus+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_mod:
  case e_int_div:
    res = checkBasicfun2Args(int_type_index, int_type_index, 
                             int_type_index, node);
    DBUG_PRINT( "tc", ("e_mod+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_real_div:
    res = checkBasicfun2Args(real_type_index, real_type_index, 
                             real_type_index, node);
    DBUG_PRINT( "tc", ("e_real_div - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_string_concat:
    res = checkBasicfun2Args(string_type_index, string_type_index, 
                             string_type_index, node);
    DBUG_PRINT( "tc", ("e_string_concat - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_prepend:
    tmp_type1 = checkBody(LEFT_OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }
    tmp_type3 = getListtypeForItemtype(tmp_type1);
    if (tmp_type3 <= 0) {
      tmp_type3 = addNotExistingListtype(tmp_type1);
    }
    res = getCompatibleType(tmp_type3, tmp_type2);
    if (res == 0) {
      tc_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
	       COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
	       tmp_type3, tmp_type2);
      //res = 0;
    }

    DBUG_PRINT( "tc", ("e_prepend - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_list_concat:
    tmp_type1 = checkBody(LEFT_OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(RIGHT_OPERAND_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }

    if (getItemtypeOfList(tmp_type1) == 0) {
      expectedX_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		      "any LIST-type", tmp_type1);
      res = 0;
    } else {
      res = getCompatibleType(tmp_type1, tmp_type2);
      if(res == 0) {
	tc_error(LINENO_NODE(RIGHT_OPERAND_EXPR_NODE(node)), 
		 COLNO_NODE( RIGHT_OPERAND_EXPR_NODE(node)),
		 tmp_type1, tmp_type2);
	res = 0;
      }
    }
    DBUG_PRINT( "tc", ("e_list_concat - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_abs:
    tmp_type1 = checkBody(OPERAND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    if ( (!isFoundTypeCompatibleToExpected(int_type_index,  tmp_type1)) && 
	 (!isFoundTypeCompatibleToExpected(real_type_index, tmp_type1))   ) {
      expectedX_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( OPERAND_EXPR_NODE(node)),
		      "int, real", tmp_type1);
      res = 0;
    } else {
      res = tmp_type1;
    }
    DBUG_PRINT( "tc", ("e_abs - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_min:
  case e_max:
    res = checkBasicfun2Args(int_type_index, int_type_index, 
			     int_type_index, node); 
    DBUG_PRINT( "tc", ("e_min+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_sqrt:
  case e_sin: 
  case e_cos:
  case e_exp:
  case e_ln:
    res = checkBasicfun1Arg(real_type_index, real_type_index, node); 
    DBUG_PRINT( "tc", ("e_sqrt+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_real:
    res = checkBasicfun1Arg(int_type_index, real_type_index, node); 
    DBUG_PRINT( "tc", ("e_real - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_truncate:
    res = checkBasicfun1Arg(real_type_index, int_type_index, node); 
    DBUG_PRINT( "tc", ("e_truncate - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_size:
    res = checkBasicfun1Arg(string_type_index, int_type_index, node); 
    DBUG_PRINT( "tc", ("e_size - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_ord:
    res = checkBasicfun1Arg(string_type_index, int_type_index, node); 
    DBUG_PRINT( "tc", ("e_ord - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_chr:
    res = checkBasicfun1Arg(int_type_index, string_type_index, node); 
    DBUG_PRINT( "tc", ("e_chr - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_substring:
    tmp_type1 = checkBody(COND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(THEN_EXPR_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }
    tmp_type3 = checkBody(ELSE_EXPR_EXPR_NODE(node));
    if (tmp_type3 == 0) { DBUG_RETURN(0); }

    if        (!isFoundTypeCompatibleToExpected(string_type_index, tmp_type1)){
      tc_error(LINENO_NODE(COND_EXPR_NODE(node)), 
	       COLNO_NODE( COND_EXPR_NODE(node)),
	       bool_type_index, tmp_type1);
      res = 0;
    } else if (!isFoundTypeCompatibleToExpected(int_type_index, tmp_type2)) {
      tc_error(LINENO_NODE(THEN_EXPR_EXPR_NODE(node)), 
	       COLNO_NODE( THEN_EXPR_EXPR_NODE(node)),
	       int_type_index, tmp_type2);
      res = 0;
    } else if (!isFoundTypeCompatibleToExpected(int_type_index, tmp_type3)) {
      tc_error(LINENO_NODE(ELSE_EXPR_EXPR_NODE(node)), 
	       COLNO_NODE( ELSE_EXPR_EXPR_NODE(node)),
	       int_type_index, tmp_type3);
      res = 0;
    } else {
      res = string_type_index;
    }
    DBUG_PRINT( "tc", ("e_substring - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_null:
    tmp_type1 = checkBody(OPERAND_EXPR_NODE(node));
    if (getItemtypeOfList(tmp_type1) == 0) {
      expectedX_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( OPERAND_EXPR_NODE(node)),
		      "any LIST-type", tmp_type1);
      res = 0;
    } else {
      res = bool_type_index;
    }
    DBUG_PRINT( "tc", ("e_null - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_hd:
    tmp_type1 = checkBody(OPERAND_EXPR_NODE(node));
    tmp_type3 = getItemtypeOfList(tmp_type1);
    if (tmp_type3 == 0) {
      expectedX_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( OPERAND_EXPR_NODE(node)),
		      "any LIST-type", tmp_type1);
      res = 0;
    } else {
      res = tmp_type3;
    }
    DBUG_PRINT( "tc", ("e_hd - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_length:
    tmp_type1 = checkBody(OPERAND_EXPR_NODE(node));
    tmp_type3 = getItemtypeOfList(tmp_type1);
    if (tmp_type3 == 0) {
      expectedX_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( OPERAND_EXPR_NODE(node)),
		      "any LIST-type", tmp_type1);
      res = 0;
    } else {
      res = int_type_index;
    }
    DBUG_PRINT( "tc", ("e_length - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_tl:
  case e_rev:
    tmp_type1 = checkBody(OPERAND_EXPR_NODE(node));
    if (getItemtypeOfList(tmp_type1) == 0) {
      expectedX_error(LINENO_NODE(OPERAND_EXPR_NODE(node)), 
		      COLNO_NODE( OPERAND_EXPR_NODE(node)),
		      "any LIST-type", tmp_type1);
      res = 0;
    } else {
      res = tmp_type1;
    }
    DBUG_PRINT( "tc", ("e_tl+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_ifthenelse:
    tmp_type1 = checkBody(COND_EXPR_NODE(node));
    if (tmp_type1 == 0) { DBUG_RETURN(0); }
    tmp_type2 = checkBody(THEN_EXPR_EXPR_NODE(node));
    if (tmp_type2 == 0) { DBUG_RETURN(0); }
    tmp_type3 = checkBody(ELSE_EXPR_EXPR_NODE(node));
    if (tmp_type3 == 0) { DBUG_RETURN(0); }

    if (!isFoundTypeCompatibleToExpected(bool_type_index, tmp_type1)) {
      tc_error(LINENO_NODE(COND_EXPR_NODE(node)), 
	       COLNO_NODE( COND_EXPR_NODE(node)),
	       bool_type_index, tmp_type1);
      res = 0;
    } else {
      res = getCompatibleType(tmp_type2, tmp_type3);
      if (res == 0) {
	difftypes_error(LINENO_NODE(THEN_EXPR_EXPR_NODE(node)), 
			COLNO_NODE( THEN_EXPR_EXPR_NODE(node)),
			tmp_type2, tmp_type3);
	//res = 0;
      }
    }
    DBUG_PRINT( "tc", ("e_ifthenelse - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  case e_constructor:
  case e_userfun:
    res = checkUserfun(node);
    DBUG_PRINT( "tc", ("e_constructor+ - %s (%d)", seekSymbolTypeindex(
                       symbol_table, res)->id, 
                       res) );
    break;

  default:
    printf_expr_node(node);
    tc_error(LINENO_NODE(node), COLNO_NODE(node), 0, 0);
    res = 0;
  } //switch

  node->valuetype = res;
  //printf("leaving checkBody: res = %d\n", res);
  DBUG_RETURN(res);
} //checkBody


