#include <stdio.h>
#include <gtk/gtk.h>
#include <string.h>
#include "../Misc/my_dbug.h"
#include "../Misc/error_interface.h"
#include "../Syntaxtree/symboltable.h"
#include "../Syntaxtree/syntaxtree.h"
#include "resources.h"
#include "tc_util.h"


MLType unit_type_index, bool_type_index, int_type_index, 
       real_type_index, string_type_index, nil_type_index, list_type_index;
GList* act_binding;




/** Seeks through the symbol_table for a entry where the given listtype is
  * typeindex and the symboltype is e_st_basictypr or e_st_datatype.
  * It returnes the typeindex of the listitems.
  * @return -1 if listtype was not found, 0 if listtype was not a list or
  *         the typeindex (>0) of the listitems.
  */
MLType
getItemtypeOfList(MLType listtype)
{
  MLType res;
  Symboltable_entry_ptr pSte;
  
  
  DBUG_ENTER("getItemtypeOfList");

  pSte = seekSymbolTypeindex(symbol_table, listtype);
  if (pSte != NULL) {
    //if it is not a listype then has to be type_of_listitems 0
    res = pSte->type_of_listitems;
  } else {
    res = -1;
    DBUG_PRINT("seek",("listtype %d not found in symbol_table", listtype));
  }
  
  DBUG_RETURN(res);
} //getItemtypeOfList


/** Seeks through the symbol_table for a entry where the given itemtype is
  * type_of_listitems and the symboltype is e_st_basictypr.
  * It returnes the type_index of the list.
  * @return -1 if itemtype was not found, or the typeindex (>0) 
  *         of the listitems.
  */
MLType
getListtypeForItemtype(MLType itemtype)
{
  MLType res;
  Symboltable_entry_ptr pSte;
  
  
  DBUG_ENTER("getListtypeForItemtype");

  pSte = seekSymbolTypeOfListitems(symbol_table, itemtype);
  if (pSte != NULL) {
    res = pSte->type_index;
  } else {
    res = -1;
    DBUG_PRINT("seek",("itemtype %d not found in symbol_table", itemtype));
  }
  
  DBUG_RETURN(res);
} //getListtypeForItemtype


static void
error_cleanup(void)
{
  DBUG_ENTER("error_cleanup");

  //clean up so things (e.g. free mem)
  //syntax_tree       = free_syntaxtree(syntax_tree);
  //symbol_table      = free_symboltable(symbol_table);

  DBUG_VOID_RETURN;
} //error_cleanup


/** TypeChecking_ERROR is a error handling function for this c-file 
  */
void
tc_error(gint lineno, gint colno, MLType expected, MLType found)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];
  Symboltable_entry_ptr pSteE, pSteF;
  gchar *strE, *strF;

  DBUG_ENTER("tc_error");

  pSteE = seekSymbolTypeindex(symbol_table, expected);
  pSteF = seekSymbolTypeindex(symbol_table, found);
  if (pSteE != NULL) { strE = pSteE->id; } 
  else               { strE = ""; }
  if (pSteF != NULL) { strF = pSteF->id; } 
  else               { strF = ""; }

  sprintf(line0, "Type mismatch");
  sprintf(line1, "-------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "Expected type: %s (%d)", strE, expected);
  sprintf(line4, "Found type   : %s (%d)", strF, found);
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //tc_error


/** TypeChecking_ERROR is a error handling function for this c-file 
  */
void
difftypes_error(gint lineno, gint colno, MLType left, MLType right)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];
  Symboltable_entry_ptr pSteL, pSteR;
  gchar *strL, *strR;

  DBUG_ENTER("difftypes_error");

  pSteL = seekSymbolTypeindex(symbol_table, left);
  pSteR = seekSymbolTypeindex(symbol_table, right);
  if (pSteL != NULL) { strL = pSteL->id; } 
  else               { strL = ""; }
  if (pSteR != NULL) { strR = pSteR->id; } 
  else               { strR = ""; }

  sprintf(line0, "Different types for operand");
  sprintf(line1, "---------------------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "left type : %s (%d)", strL, left);
  sprintf(line4, "right type: %s (%d)", strR, right);
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //difftypes_error


void
foundIntOrReal_error(gint lineno, gint colno, MLType expected)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];
  Symboltable_entry_ptr pSteE;
  gchar *strE;

  DBUG_ENTER("foundIntOrReal_error");

  pSteE = seekSymbolTypeindex(symbol_table, expected);
  if (pSteE != NULL) { strE = pSteE->id; } 
  else               { strE = ""; }

  sprintf(line0, "Type mismatch");
  sprintf(line1, "-------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "Expected type: %s (%d)", strE, expected);
  sprintf(line4, "Found type   : int or real");
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //foundIntOrReal_error


void
expectedX_error(gint lineno, gint colno, gchar* expectedStr, MLType found)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];
  Symboltable_entry_ptr pSteE;
  gchar *strE;

  DBUG_ENTER("expectedX_error");

  pSteE = seekSymbolTypeindex(symbol_table, found);
  if (pSteE != NULL) { strE = pSteE->id; } 
  else               { strE = ""; }

  sprintf(line0, "Type mismatch");
  sprintf(line1, "-------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "Expected type: %s", expectedStr);
  sprintf(line4, "Found type   : %s (%d)", strE, found);
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //expectedX_error


/** ParameterCount_ERROR is a error handling function for this c-file 
  */
void
pc_error(gint lineno, gint colno, gint expected, gint found)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];

  DBUG_ENTER("pc_error");

  sprintf(line0, "Wrong number of parameters");
  sprintf(line1, "--------------------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "Expected parameters: %d", expected);
  sprintf(line4, "Found parameters   : %d", found);
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //pc_error


/** ParameterCount_ERROR is a error handling function for this c-file 
  */
void
let_pc_error(gint lineno, gint colno)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];

  DBUG_ENTER("pc_error");

  sprintf(line0, "Wrong number of variables");
  sprintf(line1, "--------------------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "number of variables doesn't");
  sprintf(line4, "match the product-type.");
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //pc_error


/** ParameterCount_ERROR is a error handling function for this c-file 
 */
void
    project_error(gint lineno, gint colno)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
  line5[c_max];

  DBUG_ENTER("pc_error");

  sprintf(line0, "Projection-Error");
  sprintf(line1, "--------------------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "first parameter must be constant,");
  sprintf(line4, "second parameter must have enough parts.");
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  error_cleanup();

  DBUG_VOID_RETURN;
} //pc_error

gboolean
isFoundTypeCompatibleToExpected(MLType expected, MLType found)
{
  gboolean res;
  Symboltable_entry_ptr pSteE, pSteF;

  DBUG_ENTER("isFoundTypeCompatibleToExpected");

  if (expected == found) {
    res = TRUE;
  } else {
    pSteE = seekSymbolTypeindex(symbol_table, expected);
    DBUG_ASSERT(pSteE != NULL);
    pSteF = seekSymbolTypeindex(symbol_table, found);
    DBUG_ASSERT(pSteF != NULL);

    if (    pSteE->symbol_type == e_st_product 
         && pSteF->symbol_type == e_st_product ) { 
	Product_part_ptr ppE = pSteE->product_parts;
	Product_part_ptr ppF = pSteF->product_parts;
	res = TRUE;
	while ( ppE != NULL && ppF != NULL ) {
		if ( ! isFoundTypeCompatibleToExpected(ppE->type, ppF->type) ) {	
			res = FALSE;
			ppE = NULL;
		} else {
			ppE = ppE->next;
			ppF = ppF->next;	
		}
	}
	if ( ppE != NULL || ppF != NULL ) {
		res = FALSE;	
	}
    } else {
      //if (pSteF->id == "list" or starts with "list ") and ...
      if ( (0 == strncmp(pSteF->id, "list", 4)) 
	 && ( (pSteF->id[4] == '\0') || (pSteF->id[4] == ' ') ) 
	 && (SYMBOL_LISTNESTING(pSteE) >= SYMBOL_LISTNESTING(pSteF)) ){
        res = TRUE;
      } else {
        res = FALSE;
      } 
    }
  }

  DBUG_RETURN(res);
} //isFoundTypeCompatibleToExpected


/** Returns the compatible type-index if both types are equal. If both types
  * are lists it returns the correct type even if one is a (even nested) 
  * nil-list.
  * @return 0 in case of incompatible types and the compatible type-index 
  *        if both are compatible  
  */
MLType
getCompatibleType(MLType t1, MLType t2)
{
  MLType res;
  Symboltable_entry_ptr pSte1, pSte2;

  DBUG_ENTER("getCompatibleType");

  if        (t1 == t2) {
    res = t1;
  } else {
    pSte1 = seekSymbolTypeindex(symbol_table, t1);
    DBUG_ASSERT(pSte1 != NULL);
    pSte2 = seekSymbolTypeindex(symbol_table, t2);
    DBUG_ASSERT(pSte2 != NULL);
    g_print("estpr %i", e_st_product);
    g_print("sssssss %i", pSte2->symbol_type);
    // if both are product types, the check all parts(mem_management_0)
    if ( e_st_product  == pSte1->symbol_type &&  e_st_product  == pSte2->symbol_type ) {
      return t1;
           //if (pSte1->id == "list" or starts with "list ") and ...
    } else  if        ( (0 == strncmp(pSte1->id, "list", 4)) 
	        && ((pSte1->id[4] == '\0') || (pSte1->id[4] == ' ')) 
	        && (SYMBOL_LISTNESTING(pSte1) <= SYMBOL_LISTNESTING(pSte2)) ) {
      res = t2;
           //if (pSte2->id == "list" or starts with "list ") and ...
    } else if ( (0 == strncmp(pSte2->id, "list", 4)) 
		&& ((pSte2->id[4] == '\0') || (pSte2->id[4] == ' ')) 
		&& (SYMBOL_LISTNESTING(pSte1) >= SYMBOL_LISTNESTING(pSte2)) ) {
      res = t1;
    } else {
      res = 0;
    } //if
  }

  DBUG_RETURN(res);
} //getCompatibleType


